@@ -1,23 +1,11 @@
 import functools
-import json
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Iterable,
-    Optional,
-    Sequence,
-    Set,
-    Tuple,
-    Union,
-    overload,
-)
+import typing as t
 
-from flask import Blueprint, Response, blueprints, current_app, g, jsonify, request
-from jsonschema.validators import Draft7Validator
-from werkzeug.exceptions import BadRequest, NotFound, NotImplemented, Unauthorized
+from flask import Blueprint, blueprints, current_app, g, jsonify, request
+from pydantic import ValidationError
+from werkzeug.exceptions import BadRequest as WerkzeugBadRequest
 
-from globus_action_provider_tools.authentication import TokenChecker
+from globus_action_provider_tools.authentication import AuthState, TokenChecker
 from globus_action_provider_tools.authorization import (
     authorize_action_access_or_404,
     authorize_action_management_or_404,
@@ -25,35 +13,35 @@ from globus_action_provider_tools.authorization import (
 from globus_action_provider_tools.data_types import (
     ActionProviderDescription,
     ActionProviderJsonEncoder,
-    ActionRequest,
     ActionStatus,
     ActionStatusValue,
-    AuthState,
 )
-from globus_action_provider_tools.flask import (
-    blueprint_error_handler,
-    flask_validate_request,
+from globus_action_provider_tools.flask.exceptions import (
+    ActionNotFound,
+    ActionProviderError,
+    BadActionRequest,
+    UnauthorizedRequest,
 )
 from globus_action_provider_tools.flask.helpers import (
+    action_status_return_to_view_return,
+    blueprint_error_handler,
+    check_token,
+    get_input_body_validator,
     parse_query_args,
     query_args_to_enum,
+    validate_input,
+)
+from globus_action_provider_tools.flask.types import (
+    ActionCancelType,
+    ActionEnumerationType,
+    ActionLoaderType,
+    ActionLogType,
+    ActionRunType,
+    ActionSaverType,
+    ActionStatusReturn,
+    ActionStatusType,
+    ViewReturn,
 )
-
-ActionStatusReturn = Union[ActionStatus, Tuple[ActionStatus, int]]
-ActionStatusType = Callable[[Union[str, ActionStatus], AuthState], ActionStatusReturn]
-
-ActionLogReturn = Dict[str, Any]
-ActionLogType = Callable[[str, AuthState], ActionLogReturn]
-
-ActionRunType = Callable[[ActionRequest, AuthState], ActionStatusReturn]
-ActionCancelType = ActionStatusType
-ActionReleaseType = ActionStatusType
-ActionEnumerationType = Callable[[AuthState, Dict[str, Set]], Sequence[ActionStatus]]
-
-ViewReturn = Union[Tuple[Response, int], Tuple[str, int]]
-
-ActionLoaderType = Tuple[Callable[[str, Any], ActionStatus], Any]
-ActionSaverType = Tuple[Callable[[ActionStatus, Any], None], Any]
 
 
 class ActionProviderBlueprint(Blueprint):
@@ -61,8 +49,8 @@ class ActionProviderBlueprint(Blueprint):
         self,
         provider_description: ActionProviderDescription,
         *args,
-        globus_auth_client_name: Optional[str] = None,
-        additional_scopes: Iterable[str] = (),
+        globus_auth_client_name: t.Optional[str] = None,
+        additional_scopes: t.Iterable[str] = (),
         **kwarg,
     ):
         """Create a new ActionProviderBlueprint. All arguments not listed here are the
@@ -87,11 +75,11 @@ class ActionProviderBlueprint(Blueprint):
 
         super().__init__(*args, **kwarg)
 
-        self.action_loader_plugin: Optional[ActionLoaderType] = None
-        self.action_saver_plugin: Optional[ActionSaverType] = None
+        self.action_loader_plugin: t.Optional[ActionLoaderType] = None
+        self.action_saver_plugin: t.Optional[ActionSaverType] = None
 
         self.provider_description = provider_description
-        self.input_body_validator = self._load_input_body_validator()
+        self.input_body_validator = get_input_body_validator(provider_description)
         self.globus_auth_client_name = globus_auth_client_name
         self.additional_scopes = additional_scopes
 
@@ -101,7 +89,11 @@ class ActionProviderBlueprint(Blueprint):
         self.record_once(self._create_token_checker)
 
         self.add_url_rule(
-            "", "action_introspect", self._action_introspect, methods=["GET"]
+            "/",
+            "action_introspect",
+            self._action_introspect,
+            methods=["GET"],
+            strict_slashes=False,
         )
 
         # If using an action-loader, it's possible that the status and cancel
@@ -164,9 +156,9 @@ class ActionProviderBlueprint(Blueprint):
             allow_all_authenticated_users=True,
         ):
             current_app.logger.info(
-                f'User "{g.auth_state.effective_identity}" is unauthorized to introspect Action Provider'
+                f"{g.auth_state.effective_identity} is unauthorized to introspect Action Provider"
             )
-            raise NotFound
+            raise UnauthorizedRequest
 
         return jsonify(self.provider_description), 200
 
@@ -184,9 +176,10 @@ class ActionProviderBlueprint(Blueprint):
                 allow_all_authenticated_users=True,
             ):
                 current_app.logger.info(
-                    f'User "{g.auth_state.effective_identity}" is unauthorized to enumerate Actions'
+                    f"{g.auth_state.effective_identity} is unauthorized to enumerate "
+                    "Actions"
                 )
-                raise NotFound
+                raise UnauthorizedRequest
 
             valid_statuses = set(e.name.casefold() for e in ActionStatusValue)
             statuses = parse_query_args(
@@ -214,7 +207,7 @@ class ActionProviderBlueprint(Blueprint):
         )
         return wrapper
 
-    def action_run(self, func: ActionRunType) -> Callable[[], ViewReturn]:
+    def action_run(self, func: ActionRunType) -> t.Callable[[], ViewReturn]:
         """
         Decorates a function to be run as an Action Provider's run endpoint.
         """
@@ -226,31 +219,94 @@ class ActionProviderBlueprint(Blueprint):
                 allow_all_authenticated_users=True,
             ):
                 current_app.logger.info(
-                    f'User "{g.auth_state.effective_identity}" is unauthorized to run Action'
+                    f"{g.auth_state.effective_identity} is unauthorized to run Action"
                 )
-                raise Unauthorized()
+                raise UnauthorizedRequest
 
-            # Ensure incoming request conforms to ActionRequest schema
-            result = flask_validate_request(request)
-            if result.error_msg:
-                raise BadRequest(result.error_msg)
+            try:
+                json_input = request.get_json(force=True)
+            except WerkzeugBadRequest:
+                current_app.logger.info(
+                    f"{g.auth_state.effective_identity} submitted input that could not "
+                    f"be parsed as JSON: {str(request.data)}"
+                )
+                raise BadActionRequest("Invalid JSON")
+            try:
+                action_request = validate_input(json_input, self.input_body_validator)
+            except BadActionRequest as err:
+                current_app.logger.info(
+                    f"{g.auth_state.effective_identity} submitted invalid input: "
+                    f"{err.get_body()}"
+                )
+                raise
 
-            request_json = request.get_json(force=True)
-            action_request = ActionRequest(**request_json)
+            # It's possible the user will attempt to make a malformed ActionStatus -
+            # pydantic won't like that. So log and handle the error with a 500
+            try:
+                status = func(action_request, g.auth_state)
+            except ValidationError as ve:
+                current_app.logger.error(
+                    f"ActionProvider attempted to create a non-conformant ActionStatus "
+                    f"in {func.__name__}: {ve.errors()}"
+                )
+                raise ActionProviderError
 
-            # Ensure incoming Action body conforms to Action Provider schema
-            self._validate_input(action_request.body)
-            status = func(action_request, g.auth_state)
             self._save_action(status)
-            return self._action_status_return_to_view_return(status, 202)
+            return action_status_return_to_view_return(status, 202)
 
         # Add new and old-style AP API endpoints
         self.add_url_rule("/run", None, wrapper, methods=["POST"])
         self.add_url_rule("/actions", func.__name__, wrapper, methods=["POST"])
         return wrapper
 
-    @overload
-    def action_status(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+    @t.overload
+    def action_resume(self, func: t.Callable[[str, AuthState], ActionStatusReturn]):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_resume can accept a str or ActionStatus as the first arg type
+        NOTE: typing_extensions.Protocol would be better if not for it's poor
+        error messages
+        """
+        ...
+
+    @t.overload
+    def action_resume(
+        self, func: t.Callable[[ActionStatus, AuthState], ActionStatusReturn]
+    ):
+        """
+        Using these stubs w/ @overload tells mypy that the actual implementation
+        for action_resume can accept a str or ActionStatus as the first arg type
+        """
+        ...
+
+    def action_resume(self, func) -> t.Callable[[str], ViewReturn]:
+        """
+        Decorates a function to be run as an Action Provider's resume endpoint.
+        """
+
+        @functools.wraps(func)
+        def wrapper(action_id: str) -> ViewReturn:
+            # Attempt to use a user-defined function to lookup the Action based
+            # on its action_id. If an action is found, authorize access to it
+            if self.action_loader_plugin:
+                action = self._load_action_by_id(action_id)
+                authorize_action_access_or_404(action, g.auth_state)
+
+            status = func(action_id, g.auth_state)
+            return action_status_return_to_view_return(status, 200)
+
+        # Add new and old-style AP API endpoints
+        self.add_url_rule("/<string:action_id>/resume", None, wrapper, methods=["POST"])
+        self.add_url_rule(
+            "/actions/<string:action_id>/resume",
+            "action_resume",
+            wrapper,
+            methods=["POST"],
+        )
+        return wrapper
+
+    @t.overload
+    def action_status(self, func: t.Callable[[str, AuthState], ActionStatusReturn]):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
         for action_status can accept a str or ActionStatus as the first arg type
@@ -259,9 +315,9 @@ class ActionProviderBlueprint(Blueprint):
         """
         ...
 
-    @overload
+    @t.overload
     def action_status(
-        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+        self, func: t.Callable[[ActionStatus, AuthState], ActionStatusReturn]
     ):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
@@ -294,23 +350,32 @@ class ActionProviderBlueprint(Blueprint):
                 # in which case just return the bare action as the result
                 result = action
         else:
+            if not hasattr(self, "_action_status"):
+                current_app.logger.error("ActionProvider has no action status endpoint")
+                raise ActionProviderError
             try:
+                # It's possible the user will attempt to make a malformed ActionStatus -
+                # pydantic won't like that. So handle the error with a 500
                 result = self._action_status(action_id, g.auth_state)
-            except AttributeError:
-                raise NotImplemented("No status endpoint is available")
-        return self._action_status_return_to_view_return(result, 200)
+            except ValidationError as ve:
+                current_app.logger.error(
+                    f"ActionProvider attempted to create a non-conformant ActionStatus "
+                    f"in {self._action_status.__name__}: {ve.errors()}"
+                )
+                raise ActionProviderError
+        return action_status_return_to_view_return(result, 200)
 
-    @overload
-    def action_cancel(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+    @t.overload
+    def action_cancel(self, func: t.Callable[[str, AuthState], ActionStatusReturn]):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
         for action_cancel can accept a str or ActionStatus as the first arg type
         """
         ...
 
-    @overload
+    @t.overload
     def action_cancel(
-        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+        self, func: t.Callable[[ActionStatus, AuthState], ActionStatusReturn]
     ):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
@@ -344,23 +409,31 @@ class ActionProviderBlueprint(Blueprint):
             finally:
                 self._save_action(action)
         else:
+            if not hasattr(self, "_action_cancel"):
+                current_app.logger.error("ActionProvider has no action cancel endpoint")
+                raise ActionProviderError
+
             try:
                 result = self._action_cancel(action_id, g.auth_state)
-            except AttributeError:
-                raise NotImplemented("No cancel endpoint is available")
-        return self._action_status_return_to_view_return(result, 200)
+            except ValidationError as ve:
+                current_app.logger.error(
+                    f"ActionProvider attempted to create a non-conformant ActionStatus "
+                    f"in {self._action_cancel.__name__}: {ve.errors()}"
+                )
+                raise ActionProviderError
+        return action_status_return_to_view_return(result, 200)
 
-    @overload
-    def action_release(self, func: Callable[[str, AuthState], ActionStatusReturn]):
+    @t.overload
+    def action_release(self, func: t.Callable[[str, AuthState], ActionStatusReturn]):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
         for action_release can accept a str or ActionStatus as the first arg type
         """
         ...
 
-    @overload
+    @t.overload
     def action_release(
-        self, func: Callable[[ActionStatus, AuthState], ActionStatusReturn]
+        self, func: t.Callable[[ActionStatus, AuthState], ActionStatusReturn]
     ):
         """
         Using these stubs w/ @overload tells mypy that the actual implementation
@@ -368,7 +441,7 @@ class ActionProviderBlueprint(Blueprint):
         """
         ...
 
-    def action_release(self, func) -> Callable[[str], ViewReturn]:
+    def action_release(self, func) -> t.Callable[[str], ViewReturn]:
         """
         Decorates a function to be run as an Action Provider's release endpoint.
         """
@@ -381,8 +454,16 @@ class ActionProviderBlueprint(Blueprint):
                 action = self._load_action_by_id(action_id)
                 authorize_action_management_or_404(action, g.auth_state)
 
-            status = func(action_id, g.auth_state)  # type: ignore
-            return self._action_status_return_to_view_return(status, 200)
+            try:
+                status = func(action_id, g.auth_state)
+            except ValidationError as ve:
+                current_app.logger.error(
+                    f"ActionProvider attempted to create a non-conformant ActionStatus "
+                    f"in {func.__name__}: {ve.errors()}"
+                )
+                raise ActionProviderError
+
+            return action_status_return_to_view_return(status, 200)
 
         # Add new and old-style AP API endpoints
         self.add_url_rule(
@@ -399,7 +480,7 @@ class ActionProviderBlueprint(Blueprint):
         )
         return wrapper
 
-    def action_log(self, func: ActionLogType) -> Callable[[str], ViewReturn]:
+    def action_log(self, func: ActionLogType) -> t.Callable[[str], ViewReturn]:
         """
         Decorates a function to be run an an Action Provider's logging endpoint.
         """
@@ -422,7 +503,7 @@ class ActionProviderBlueprint(Blueprint):
         )
         return wrapper
 
-    def register_action_loader(self, storage_backend: Any):
+    def register_action_loader(self, storage_backend: t.Any):
         """
         Decorates a function that will be used to lookup an ActionStatus by its
         action_id. Multiple action_loaders with different backends can be
@@ -447,13 +528,13 @@ class ActionProviderBlueprint(Blueprint):
         assume the ActionStatus is non-recoverable.
         """
         if self.action_loader_plugin is None:
-            raise NotFound
+            raise ActionNotFound
 
         func, backend = self.action_loader_plugin
         action = func(action_id, backend)
         if action:
             return action
-        raise NotFound
+        raise ActionNotFound
 
     def register_action_saver(self, storage_backend):
         """
@@ -492,67 +573,4 @@ class ActionProviderBlueprint(Blueprint):
         Parses a token from a request to generate an auth_state object which is
         then made available as the second argument to decorated functions.
         """
-        access_token = (
-            request.headers.get("Authorization", "").strip().lstrip("Bearer ")
-        )
-        auth_state = self.checker.check_token(access_token)
-        g.auth_state = auth_state
-
-    def _load_input_body_validator(self) -> Union[Draft7Validator, None]:
-        """
-        Creates a JSON Validator object to be used to verify that the body of a
-        RUN request conforms to the Action Provider's defined schema. In the
-        event that no schema was provided when defining the Blueprint, no
-        validator is created and no validation occurs.
-        """
-        if isinstance(self.provider_description.input_schema, str):
-            input_schema = json.loads(self.provider_description.input_schema)
-        else:
-            input_schema = self.provider_description.input_schema
-
-        if input_schema is not None:
-            input_body_validator = Draft7Validator(input_schema)
-        else:
-            input_body_validator = None
-
-        return input_body_validator
-
-    def _validate_input(self, input: Dict[str, Any]) -> None:
-        """
-        Use a created JSON Validator to verify the input body of an incoming
-        request conforms to the defined JSON schema. In the event that the
-        validation reports any errors, a BadRequest exception gets raised.
-        """
-        if self.input_body_validator is None:
-            return
-
-        errors = self.input_body_validator.iter_errors(input)
-        error_messages = []
-        for error in errors:
-            if error.path:
-                # Elements of the error path may be integers or other non-string types,
-                # but we need strings for use with join()
-                error_path_for_message = ".".join([str(x) for x in error.path])
-                error_message = (
-                    f"'{error_path_for_message}' invalid due to {error.message}"
-                )
-            else:
-                error_message = error.message
-            error_messages.append(error_message)
-
-        if error_messages:
-            message = "; ".join(error_messages)
-            raise BadRequest(message)
-
-    def _action_status_return_to_view_return(
-        self, status: ActionStatusReturn, default_status_code: int
-    ) -> ViewReturn:
-        """
-        Helper function to return a ActionStatusReturn object as a valid Flask
-        response.
-        """
-        if isinstance(status, ActionStatus):
-            status_code = default_status_code
-        elif isinstance(status, tuple):
-            status, status_code = status
-        return jsonify(status), status_code
+        g.auth_state = check_token(request, self.checker)
